-------------------------------------------------------------------------------
--! @file       CipherCore.vhd
--! @brief      Cipher core template module
--! @project    CAESAR Candidate Evaluation
--! @author     Ekawat (ice) Homsirikamol
--! @copyright  Copyright (c) 2015 Cryptographic Engineering Research Group
--!             ECE Department, George Mason University Fairfax, VA, U.S.A.
--!             All rights Reserved.
--! @license    This project is released under the GNU Public License.
--!             The license and distribution terms for this file may be
--!             found in the file LICENSE in this distribution or at
--!             http://www.gnu.org/licenses/gpl-3.0.txt
--! @note       This is publicly available encryption source code that falls
--!             under the License Exception TSU (Technology and software-
--!             —unrestricted)
-------------------------------------------------------------------------------

library ieee;
use ieee.std_logic_1164.ALL;
use work.AEAD_pkg.all;

entity CipherCore is
    generic (
        --! Reset behavior
        G_ASYNC_RSTN    : boolean := False; --! Async active low reset
        --! Block size (bits)
        G_DBLK_SIZE     : integer := 128;   --! Data
        G_KEY_SIZE      : integer := 128;   --! Key
        G_TAG_SIZE      : integer := 128;   --! Tag
        --! The number of bits required to hold block size expressed in
        --! bytes = log2_ceil(G_DBLK_SIZE/8)
        G_LBS_BYTES     : integer := 4;
        --! Maximum supported AD/message/ciphertext length = 2^G_MAX_LEN-1
        G_MAX_LEN       : integer := SINGLE_PASS_MAX
    );
    port (
        --! Global
        clk             : in  std_logic;
        rst             : in  std_logic;
        --! PreProcessor (data)
        key             : in  std_logic_vector(G_KEY_SIZE       -1 downto 0);
        bdi             : in  std_logic_vector(G_DBLK_SIZE      -1 downto 0);
        --! PreProcessor (controls)
        key_ready       : out std_logic;
        key_valid       : in  std_logic;
        key_update      : in  std_logic;
        decrypt         : in  std_logic;
        bdi_ready       : out std_logic;
        bdi_valid       : in  std_logic;
        bdi_type        : in  std_logic_vector(3                -1 downto 0);
        bdi_partial     : in  std_logic;
        bdi_eot         : in  std_logic;
        bdi_eoi         : in  std_logic;
        bdi_size        : in  std_logic_vector(G_LBS_BYTES+1    -1 downto 0);
        bdi_valid_bytes : in  std_logic_vector(G_DBLK_SIZE/8    -1 downto 0);
        bdi_pad_loc     : in  std_logic_vector(G_DBLK_SIZE/8    -1 downto 0);
        --! PostProcessor
        bdo             : out std_logic_vector(G_DBLK_SIZE      -1 downto 0);
        bdo_valid       : out std_logic;
        bdo_ready       : in  std_logic;
        bdo_size        : out std_logic_vector(G_LBS_BYTES+1    -1 downto 0);
        msg_auth_done   : out std_logic;
        msg_auth_valid  : out std_logic
    );
end entity CipherCore;

architecture structure of CipherCore is

signal tag_valid, load_key, S_init : std_logic;
signal en_store_bdi, en_T_beta_prime, en_tau_alpha_valid, left_shift_select : std_logic;
signal M_select, x_sel, bdo_select, right_shift_select: std_logic_vector(1 downto 0);
signal block_count_en , load_bdi_valid_bytes, first_AD_block_init, first_AD_block_en: std_logic;
signal first_M_block_init, first_M_block_en, first_M_block : std_logic;
signal en_last, en_T : std_logic;

signal En_2iL: STD_LOGIC;   
signal en_X: STD_LOGIC;   
signal tag_processing : std_logic;
signal AES_hash_start: STD_LOGIC;   
signal AES_hash_init: STD_LOGIC;   
signal AES_hash_done: STD_LOGIC;   
signal AES_hash_done_init: STD_LOGIC;   
signal AES_hash_ready: STD_LOGIC;
signal AES_hash_sel : std_logic;   
signal AES_select: STD_LOGIC_VECTOR(2 downto 0); 
signal sel_const: STD_LOGIC_VECTOR(2 downto 0); 
signal keysel: STD_LOGIC;   
signal init_sigma: STD_LOGIC;   
signal en_sigma: STD_LOGIC;   
signal block_count_init : STD_LOGIC;
signal init_X_and_Y : STD_LOGIC;

signal AES_main_start: STD_LOGIC;   
signal AES_main_decrypt: STD_LOGIC;   
signal AES_main_init: STD_LOGIC;   
signal AES_main_done: STD_LOGIC;   
signal AES_main_done_init: STD_LOGIC;   
signal AES_main_ready: STD_LOGIC;  
signal en_Y: STD_LOGIC;   
signal en_tau: STD_LOGIC;   
signal en_K: STD_LOGIC;   
signal en_L: STD_LOGIC;   
signal en_S: STD_LOGIC;   
signal en_Kf: STD_LOGIC;   
signal en_T_alpha: STD_LOGIC;
signal en_X_decrypt : std_logic;
signal decrypt_reg : std_logic; 
signal decrypt_init : std_logic;  
signal x_decrypt_sel : std_logic;
signal no_M : std_logic;

begin

datapath: entity work.POET_datapath(Structural)
			 port map(clk => clk,
						rst => rst,
						bdi => bdi,
						key => key,
						load_key => load_key,
						tag_valid => tag_valid,
						bdo => bdo,
						bdi_size => bdi_size,
						bdi_valid_bytes => bdi_valid_bytes,
						decrypt => decrypt,
						decrypt_reg_out => decrypt_reg,
						decrypt_init => decrypt_init,
						En_2iL => En_2iL,
						en_X => en_X,
						x_decrypt_sel => x_decrypt_sel,
						en_X_decrypt => en_X_decrypt,
						tag_processing => tag_processing,
						AES_hash_start => AES_hash_start,
						AES_hash_init => AES_hash_init,
						AES_hash_done_init => AES_hash_done_init,
						AES_hash_ready => AES_hash_ready,
						AES_hash_sel => AES_hash_sel,
						AES_select => AES_select,
						sel_const => sel_const,
						keysel => keysel,
						init_sigma => init_sigma,
						en_sigma => en_sigma,
						block_count_en => block_count_en,
						block_count_init => block_count_init,
						init_X_and_Y => init_X_and_Y,
						AES_main_start => AES_main_start,
						AES_main_decrypt => AES_main_decrypt,
						AES_main_init => AES_main_init,
						AES_main_done => AES_main_done,
						AES_main_done_init => AES_main_done_init,
						AES_main_ready => AES_main_ready,
						en_Y => en_Y,
						en_tau => en_tau,
						en_K => en_K,
						en_L => en_L,
						en_S => en_S,
						en_Kf => en_Kf,
						en_T_alpha => en_T_alpha,
						en_T => en_T,
						en_T_beta_prime => en_T_beta_prime,
						en_tau_alpha_valid => en_tau_alpha_valid,
						en_store_bdi => en_store_bdi,
						en_last => en_last,
						no_M => no_M,
						
						S_init => S_init, 
						M_select => M_select, 
						x_sel => x_sel,
						bdo_select => bdo_select,
						right_shift_select => right_shift_select,
						left_shift_select => left_shift_select,
						load_bdi_valid_bytes => load_bdi_valid_bytes,
						first_M_block_init => first_M_block_init,
						first_M_block_en => first_M_block_en,
						first_M_block => first_M_block,
						first_AD_block_init => first_AD_block_init,
						first_AD_block_en => first_AD_block_en
										
						);				
						
controller: entity work.Controller(Behavioral)
				port map(key_update => key_update,
							key_valid => key_valid,
							key_ready => key_ready,
							load_key => load_key,
							bdi_ready => bdi_ready,
							decrypt => decrypt_reg,
							decrypt_init => decrypt_init,
							bdi_eot => bdi_eot,
							bdi_eoi => bdi_eoi,
							bdi_valid => bdi_valid,
							bdi_size => bdi_size, 
							bdi_type => bdi_type,
							bdo_ready => bdo_ready,
							bdo_valid => bdo_valid,
							tag_valid => tag_valid,
							msg_auth_valid => msg_auth_valid,
							msg_auth_done => msg_auth_done,
							clk => clk,
							rst => rst,
							tag_processing => tag_processing,
						   x_decrypt_sel => x_decrypt_sel,
							en_X_decrypt => en_X_decrypt,							
							M_select => M_select, 
							x_sel => x_sel,
							bdo_select => bdo_select,
							right_shift_select => right_shift_select,
							left_shift_select => left_shift_select,
							block_count_en => block_count_en,
							load_bdi_valid_bytes => load_bdi_valid_bytes,
							first_AD_block_init => first_AD_block_init,
							first_AD_block_en => first_AD_block_en,
							AES_hash_start => AES_hash_start,
							AES_hash_init => AES_hash_init,
							AES_hash_done_init => AES_hash_done_init,
							AES_hash_ready => AES_hash_ready,
							AES_hash_sel => AES_hash_sel,
							AES_select => AES_select,
							sel_const => sel_const,
							keysel => keysel,
						
							init_sigma => init_sigma,
							S_init => S_init, 
							block_count_init => block_count_init, 
							init_X_and_Y => init_X_and_Y,
							first_M_block_init => first_M_block_init,
						   first_M_block_en => first_M_block_en,
						   first_M_block => first_M_block,
							no_M_out => no_M,
							
							AES_main_start => AES_main_start,
							AES_main_decrypt => AES_main_decrypt,
							AES_main_init => AES_main_init,
							AES_main_done => AES_main_done,
							AES_main_done_init => AES_main_done_init,
							AES_main_ready => AES_main_ready,
							
							en_sigma => en_sigma,
							En_2iL => En_2iL,
							en_X => en_X,
							en_T_alpha => en_T_alpha,
							en_T_beta_prime => en_T_beta_prime,
							en_tau_alpha_valid => en_tau_alpha_valid,
							en_store_bdi => en_store_bdi,
							en_last => en_last,
							en_Y => en_Y,
							en_tau => en_tau,
							en_K => en_K,
							en_L => en_L,
							en_S => en_S,
							en_Kf => en_Kf,
							en_T => en_T
							); 

end structure;