
-------------------------------------------------------------------------------
--! @file       Controller.vhd
--! @author     William Diehl
--! @brief      POET Controller
--! @date       10 July 2016
-------------------------------------------------------------------------------

library IEEE;
use IEEE.STD_LOGIC_1164.ALL;


entity Controller is
	Port (	
				clk: in std_logic;
				rst: in std_logic;
				
				--EXTERNAL INTERFACE
				
				key_update: in std_logic;
				key_valid: in std_logic;
				key_ready: out std_logic;
				bdi_ready: out std_logic;
				bdi_valid: in std_logic;
				decrypt: in std_logic;
				decrypt_init: out std_logic;
				bdi_eot: in std_logic;
				bdi_eoi: in std_logic;
				bdi_type: in std_logic_vector(2 downto 0);
				bdi_size: in std_logic_vector(4 downto 0);
				
				bdo_ready: in std_logic;
				bdo_valid: out std_logic;
				msg_auth_valid: out std_logic;
				msg_auth_done: out std_logic;
								
				--INTERFACE TO DATAPATH
				
				tag_valid : in std_logic;
				load_key : out std_logic;
				init_X_and_Y: out std_logic;
				S_init : out std_logic;
				block_count_init : out std_logic;
				En_2iL: out STD_LOGIC; 
				en_X: out STD_LOGIC;  
				en_X_decrypt : out std_logic;
				en_T_beta_prime: out std_logic; 
				en_tau_alpha_valid : out std_logic;
				x_decrypt_sel : out std_logic;
				en_store_bdi: out std_logic;
				M_select : out std_logic_vector(1 downto 0);
				x_sel : out std_logic_vector(1 downto 0);
				bdo_select : out std_logic_vector(1 downto 0);
				right_shift_select : out std_logic_vector(1 downto 0);
				left_shift_select : out std_logic;
				
				block_count_en : out std_logic;
				load_bdi_valid_bytes : out std_logic;
				first_AD_block_init : out std_logic;
				first_AD_block_en : out std_logic;
				first_M_block_init : out std_logic;
				first_M_block_en : out std_logic;
				first_M_block : in std_logic;
				en_last : out std_logic;
				no_M_out : out std_logic;

				
				AES_hash_start: out STD_LOGIC;   
				AES_hash_init: out STD_LOGIC;   
				AES_hash_done_init: in STD_LOGIC;   
				AES_hash_ready: in STD_LOGIC;   
				AES_hash_sel : out std_logic;
				AES_select: out STD_LOGIC_VECTOR(2 downto 0); 
				sel_const: out STD_LOGIC_VECTOR(2 downto 0); 
				keysel: out STD_LOGIC;   
				init_sigma: out STD_LOGIC;   
				en_sigma: out STD_LOGIC;   
				AES_main_start: out STD_LOGIC;   
				AES_main_decrypt: out STD_LOGIC;   
				AES_main_init: out STD_LOGIC;   
				AES_main_done: in STD_LOGIC;   
				AES_main_done_init: in STD_LOGIC;   
				AES_main_ready: in STD_LOGIC;   
				en_Y: out STD_LOGIC;   
				en_tau: out STD_LOGIC;   
				en_K: out STD_LOGIC;   
				en_L: out STD_LOGIC;   
				en_S: out STD_LOGIC;   
				en_Kf: out STD_LOGIC;  
				tag_processing : out STD_LOGIC;
				en_T_alpha : out STD_LOGIC;
				en_T : out std_logic
				
				);   
			  
end Controller;

architecture Behavioral of Controller is

						  
	signal state : std_logic_vector(7 downto 0);
	signal state_next: std_logic_vector(7 downto 0);	
	signal no_M_en, no_M, no_M_init, first_M_block_flag : std_logic;

begin

	p_fsm: process(clk)
		 begin
			  if rising_edge(clk) then
					if (rst = '1') then
						 state <= x"00"; --S_RESET;
					else
						 state <= state_next;
					end if;
					
			  end if;
		 end process;
    
    p_comb: process(state, key_valid, key_update, AES_main_done_init, bdi_size, bdi_type,
	 AES_main_done, AES_main_ready, bdi_valid, decrypt, AES_hash_ready,
	 tag_valid, bdo_ready, bdi_eot, bdi_eoi, first_M_block, first_M_block_flag, no_M)
    begin
        --! Default values
        
			state_next  <= state;
			key_ready <= '0';
			bdi_ready <= '0';
			bdo_valid <= '0';
			msg_auth_valid <= '0';
			msg_auth_done <= '0';
			En_2iL <= '0';
			en_X <= '0';
			en_T_beta_prime <= '0';
			en_tau_alpha_valid <= '0';
			tag_processing <= '0';
			x_decrypt_sel <= '0';
			en_X_decrypt <= '0';
			load_key <= '0';
			en_last <= '0';
			en_store_bdi <= '0';
						
			AES_hash_start <= '0'; 
			AES_hash_init <= '0'; 
			AES_hash_sel <= '0';
			AES_select <= "000";
			sel_const <= "000";
			keysel <= '0';
			init_sigma <= '0';
			en_sigma <= '0';
			AES_main_start <= '0';
			AES_main_decrypt <= '0';
			AES_MAIN_init <= '0';
			S_init <= '0';
			en_Y <= '0';
			en_tau <= '0';
			en_K <= '0';
			en_L <= '0';
			en_S <= '0';
			en_Kf <= '0';
			en_T <= '0';
			en_T_alpha <= '0';
			block_count_en <= '0';
			block_count_init <= '0';
			decrypt_init <= '0';
			init_X_and_Y <= '0';
			no_M_init <= '0';
			no_M_en <= '0';
			first_AD_block_en <= '0';
			first_AD_block_init <= '0'; 
			first_M_block_init <= '0';
			first_M_block_en <= '0';
			load_bdi_valid_bytes <= '0';
							
        case state is
            when x"00" => --s_reset 
                init_sigma <= '1';  -- initialize sigma register to 0
					 En_sigma <= '1';
					 block_count_init <= '1'; -- initialize block counter to 1
					 load_bdi_valid_bytes <= '1';
					 first_AD_block_init <= '1';
					 first_M_block_init <= '1';
					 first_M_block_flag <= '0';
					 no_M_init <= '1';
					 S_init <= '1';
					 en_S <= '1';
					 en_T <= '1';
					 en_T_beta_prime <= '1';
					 state_next <= x"01"; 
					

				when x"01" => -- s_key_update
				-- Key check
				if (bdi_valid = '1') then
					if (key_update = '1') then
						if (key_valid = '1') then
							key_ready <= '1';
							load_key <= '1';
							state_next <= x"02"; -- start initialization of subkey
						else
							state_next <= x"01";
						end if;
					else 
						state_next <= x"09"; -- if no new key, no initialization required
					end if;
				else
					state_next <= x"01";
		
				end if;
				
				when x"02" => -- s_start_init
				
				AES_main_init <= '1';
				keysel <= '1';
				state_next <=x"03"; 
				
				when x"03" => --s_AES_init_SK =>
					keysel <= '1';
					if (AES_main_done_init = '1') then
						AES_select <= "010";
						AES_main_decrypt <= '0';
						AES_main_start <= '1';
						state_next <= x"04"; -- s_encrypt_K;
					else
						state_next <= x"03"; -- s_AES_init_SK;
					end if;
					
				when x"04" => --s_encrypt_K =>
					AES_select <= "010";
					keysel <= '1';
					if (AES_main_done = '1') then
						en_K <= '1';
						sel_const <= "001";
						AES_main_decrypt <= '0';
						AES_main_start <= '1';
						state_next <= x"05"; --s_encrypt_L;
					else
						state_next <= x"04"; --s_encrypt_K;
					end if;
						
				when x"05" => --s_encrypt_L =>
					AES_select <= "010";
					keysel <= '1';
					if (AES_main_done = '1') then
						en_L <= '1';
						sel_const <= "010";
						AES_main_decrypt <= '0';
						AES_main_start <= '1';
						state_next <= x"06"; -- s_encrypt_Kf
					else
						state_next <= x"05"; --s_encrypt_L;
					end if;
					
				when x"06" => -- s_encrypt_Kf =>
					AES_select <= "010";
					keysel <= '1';
					if (AES_main_done = '1') then
					   en_Kf <= '1'; 
						AES_main_decrypt <= '0';
						AES_main_init <= '1';
						keysel <= '0';
						state_next <= x"07"; -- s_AES_hash_init
					else
						state_next <= x"06"; -- s_encrypt_Kf
					end if;
									
				when x"07" => -- s_AES_hash_init
				   AES_hash_init <= '1'; -- must occur at least a clock cycle after en_Kf
					state_next <= x"08"; -- generate special header
				
				when x"08" => --s_special_header =>
								 
					if (AES_main_done_init = '1') then
						AES_select <= "111"; -- select L
					   AES_main_decrypt <= '0';
					   AES_main_start <= '1';
					   state_next <= x"0a"; -- finish_special_header
					else
						state_next <= x"08"; --s_special_header
					end if;
					
				when x"09" => -- s_gen_spec_hdr
				-- generate special header when there was no initialization

					  AES_select <= "111"; -- select L
					  AES_main_decrypt <= '0';
					  AES_main_start <= '1';
					  state_next <= x"0a"; -- finish_special_header
					
				when x"0a" => --s_finish_special_header 
				
				if (AES_main_ready = '1') then
						En_sigma <= '1';
						state_next <= x"0b"; -- s_process_AD
					else 
						state_next <= x"0a"; -- finish special header
					end if;				
				
				when x"0b" => -- s_process_AD
				
					if (bdi_valid = '1') then
						AES_select <= "011"; -- select sigma path
						AES_main_decrypt <= '0';
						AES_main_start <= '1';
						En_2iL <= '1'; 
						first_AD_block_en <= '1';
						if (bdi_eot = '1') then
							if (bdi_eoi = '1') then 
								no_M_en <= '1';
							else
								bdi_ready <= '1'; -- get a next bdi if not no_M
							end if;
							state_next <= x"0d"; -- finish final AD block
						else
							bdi_ready <= '1';
							state_next <= x"0c"; -- compute next AD block
						end if;
					else
						state_next <= x"0b"; -- wait for bdi valid
					end if;
					
					
				when x"0c" => -- s_compute_AD_block_not_final
				
					if (AES_main_ready = '1') then
						En_sigma <= '1';
						state_next <= x"0b"; 
					else
						state_next <= x"0c"; 
					end if;
							
					
				when x"0d" => -- s_finish_AD_block_final  

				   if (AES_main_ready = '1') then
					   decrypt_init <= '1';
						En_sigma <= '1';  -- add result to sigma
						state_next <= x"0e"; -- process_final_Ek
					else 
						state_next <= x"0d"; -- finish final AD block
					end if;	
					
				when x"0e" => --s_start_AD_final_Ek =>
				
				   AES_main_decrypt <= '0';
				   AES_main_start <= '1';
					AES_select <= "100"; -- select tau path
					state_next <= x"0f"; 
					
				when x"0f" => -- s_finish_AD_final_Ek
				
					if (AES_main_ready = '1') then
						en_tau <= '1'; -- tau is computed - store tau
						state_next <= x"10"; -- process M
					else
						state_next <= x"0f"; -- finish AD final Ek
					end if;
				
			   when x"10" => -- s_initialize_X_and_Y
				     x_sel <= "00";	
				     init_X_and_Y <= '1';
					  
					  En_X <= '1'; -- X = tau
					  En_Y <= '1'; -- Y = tau xor 1
					  state_next <= x"11";  

			   when x"11" => -- s_start_top_hash

			   AES_hash_start <= '1';
			   AES_hash_sel <= '0'; -- start top hash
			   state_next <= x"12"; 

			   when x"12" => -- s_complete_top_hash1

			   if (AES_hash_ready = '1') then
				if (bdi_valid = '1' or no_M = '1') then
					if (bdi_eot = '1') then
						state_next <= x"16"; 
					else 
						M_select <= "10";
						x_sel <= "01";
						en_X <= '1';
						bdi_ready <= '1';
						block_count_en <= '1';
						state_next <= x"13"; 
					end if;
				else
					state_next <= x"12"; 
				end if;
			   else 
				state_next <= x"12"; 											
			   end if;
			
			   when x"13" => -- s_start_top_hash_and_Ek
			
			   AES_hash_start <= '1';
			   AES_hash_sel <= '0';
			   AES_main_start <= '1';
			   AES_select <= "000"; -- X_reg
				AES_main_decrypt <= decrypt;
			   state_next <= x"14"; 

			   when x"14" => -- s_complete_top_hash2

			   if (AES_hash_ready = '1') then
					if (bdi_valid = '1') then
						if (bdi_eot = '1') then
							if (first_M_block = '0') then 
								first_M_block_flag <= '1';
							end if;
							state_next <= x"16"; 
						else
							M_select <= "10";
							x_sel <= "01";
							en_X <= '1';
							bdi_ready <= '1';
							block_count_en <= '1';
							AES_hash_start <= '1';
							AES_hash_sel <= '1';
							first_M_block_en <= '1';
							state_next <= x"15"; 
						end if;
					else
						state_next <= x"14"; 				
					end if;
			   else
					state_next <= x"14"; 
			   end if;

			    when x"15" => -- s_finish_bot_hash_and_main 

			    if (AES_main_ready = '1' and AES_hash_ready = '1') then
				 if (bdo_ready = '1') then
					bdo_valid <= '1';
					bdo_select <= "00";
					en_Y <= '1';
					first_M_block_flag <= '1';  
					AES_main_start <= '1';
					AES_select <= "000";
					AES_main_decrypt <= decrypt;
					AES_hash_start <= '1';
					AES_hash_sel <= '0';
					state_next <= x"14"; 	
				else
					state_next <= x"15"; 
				end if;
			    else
				state_next <= x"15";
			   end if;

			   when x"16" => -- s_eot_state

			   if (decrypt = '1') then
					if (no_M = '0') then
						en_store_bdi <= '1'; -- buffer current bdi and get expected tag
						load_bdi_valid_bytes <= '1';
					end if;
					bdi_ready <= '1';
					state_next <= x"17"; 
			   else
					if (no_M = '1') then
						M_select <= "11"; -- choose tau as input
						bdi_ready <= '1';
					else
						M_select <= "00"; -- store trunc_bdi or tau_right_shifted
					end if;
					x_sel <= "01";
					right_shift_select <= "00";
					en_X <= '1';
					en_X_decrypt <= '1';
					
					if (first_M_block_flag = '1') then
						AES_hash_start <= '1'; -- start bottom hash because we need to output a block
						AES_hash_sel <= '1';
					end if;					
					
					state_next <= x"18"; 
			   end if;
		   			   
			   when x"17" => -- s_get_exp_tag

			   if (bdi_valid = '1') then
					if (no_M = '1') then
						M_select <= "10"; -- take bdi=exp_tag=T_alpha
						bdi_ready <= '1';
					else
						M_select <= "01";
					   left_shift_select <= '1';
					   right_shift_select <= "01";
					   en_X_decrypt <= '1';
						en_T_beta_prime <= '1';
					end if;
					x_sel <= "01";
					en_X <= '1';
					
					if (first_M_block_flag = '1') then
						AES_hash_start <= '1'; -- start bottom hash because we need to output a block
						AES_hash_sel <= '1';
					end if;
					
					state_next <= x"18"; 
			   else
					state_next <= x"17"; 
			   end if;
			   
			   when x"18" => -- s_start_Ek_S

			   if (AES_main_ready = '1' and AES_hash_ready = '1') then
					
					if (first_M_block_flag = '1') then -- a result is ready to write
						if (bdo_ready = '1') then
							bdo_select <= "00";
							bdo_valid <= '1';
							AES_select <= "001"; -- compute Ek|S|
							AES_main_decrypt <= '0';
							AES_main_start <= '1';
							en_Y <= '1'; 
							state_next <= x"19"; 
						else 
							state_next <= x"18"; 
						end if;
					else -- no result is ready to write
						AES_select <= "001"; -- compute Ek|S|
						AES_main_decrypt <= '0';
						AES_main_start <= '1';
						state_next <= x"19"; 
					end if;
				else  -- (AES_main_ready ='0' or AES_hash_ready = '0')
					state_next <= x"18"; -- wait until ready
				end if;
				
			   when x"19" => -- s_store_S
			
				if (AES_main_ready = '1') then
					en_S <= '1';
					AES_select <= "101";
					AES_main_start <= '1';
					AES_main_decrypt <= decrypt;
					AES_hash_start <= '1';
					AES_hash_sel <= '1'; -- start bot hash
					state_next <= x"1a"; 
				else
					state_next <= x"19"; 
				end if;

			   when x"1a" => -- s_store_last
			
			   if (AES_main_ready = '1' and AES_hash_ready = '1') then
					en_Y <= '1';
					en_last <= '1';
					state_next <= x"1b"; 
			   else
					state_next <= x"1a"; 
			   end if;

			   when x"1b" => -- s_adjust_last_and_output
				
				if (no_M = '0') then
					if (bdo_ready = '1') then
						bdo_select <= "01"; -- last block adjusted
						bdo_valid <= '1';
						left_shift_select <= '0';
						en_T_alpha <= '1'; 
						en_tau_alpha_valid <= '1'; -- during decryption, tau alpha = tau alpha prime tested here
						tag_processing <= '1'; -- choose X(encrypt) oy Y(decrypt) xor S as input to top hash
						AES_hash_start <= '1';
						AES_hash_sel <= '0'; -- start top hash
						state_next <= x"1c"; 
					else
					   
						state_next <= x"1b"; 
					end if;
				else
					en_T_alpha <= '1';
					en_tau_alpha_valid <= '1'; -- during decryption, tau alpha = tau alpha prime tested here
					state_next <= x"1f"; -- no need to compute rest of tag since it would be wasted
				end if;

			   when x"1c" => -- s_top_hash_tag

			   if (AES_hash_ready = '1') then
					
					x_sel <= "10"; -- Ft xor tau
					en_X <= '1';
					state_next <= x"1d"; 
			   else
					state_next <= x"1c"; 
			   end if;
	
			   when x"1d" => -- s_main_tag

			   AES_main_start <= '1';
			   AES_select <= "000";
				AES_main_decrypt <= '0'; -- tag computation
				if (decrypt = '1') then
					x_decrypt_sel <= '1';
				end if;
			   AES_hash_start <= '1';
			   AES_hash_sel <= '1'; -- start bot hash
			   state_next <= x"1e"; 
			   
			   when x"1e" => -- s_store_tag
			
			   if (AES_main_ready = '1' and AES_hash_ready = '1') then
					en_T <= '1'; -- store tag
					state_next <= x"1f"; 
			   else
					state_next <= x"1e"; 
			   end if;
			
			   when x"1f" => -- s_adjust_and_verify_tag
				  
			   right_shift_select <= "10"; -- select T for right shifting
				
			   if (decrypt = '1') then
					msg_auth_done <= '1';
					if (tag_valid = '1') then
						msg_auth_valid <= '1';
						if (no_M = '0') then
							bdi_ready <= '1';
						end if;
					end if;
					state_next <= x"00";
			   else
					if (bdo_ready ='1') then
						bdo_select <= "10";
						bdo_valid <= '1';
						if (no_M = '0') then
							bdi_ready <= '1';
						end if;
						bdi_ready <= '1';
						state_next <= x"00";
					else
						state_next <= x"1f"; 
					end if;
			    end if;				
			
			when others =>
				state_next <= x"00"; --s_reset;
           
        end case;
    end process;

no_M_reg: entity work.reg1 (behavioral)

	PORT MAP(
		clk => clk,
		init => no_M_init,
		en => no_M_en,
		d => '1',
		q => no_M);

no_M_out <= no_M;
end Behavioral;

